//--------------------------------------------------------------
// File     : stm32_ub_lin_slave.c
// Datum    : 13.07.2014
// Version  : 1.1
// Autor    : UB
// EMail    : mc-4u(@)t-online.de
// Web      : www.mikrocontroller-4u.de
// CPU      : STM32F4
// IDE      : CooCox CoIDE 1.7.4
// GCC      : 4.7 2012q4
// Module   : GPIO, UART, MISC
// Funktion : LIN-Schnittstelle (Slave)
//
// Hinweise : LIN-Transceiver = MAX13020/MAX13021 (oder TJA1020)
//
//            UART-TX (PA2) an MAX13020 (TXD/Pin4)
//            UART-RX (PA3) an MAX13020 (RXD/Pin1)
//            GPIO    (PA1) an MAX13020 (NSLP/Pin2)
//--------------------------------------------------------------


//--------------------------------------------------------------
// Includes
//--------------------------------------------------------------
#include "stm32_ub_lin_slave.h"



//--------------------------------------------------------------
// Globale Variabeln
//--------------------------------------------------------------
LIN_SLAVE_t LIN_SLAVE;
LIN_FRAME_t LIN_FRAME;



//--------------------------------------------------------------
// interne Funktionen
//--------------------------------------------------------------
void p_LIN_initGPIO(void);
void p_LIN_initUART(void);
void p_LIN_initNVIC(void);
void p_LIN_aktivateTransceiver(void);
uint8_t p_LIN_makeChecksum(LIN_FRAME_t *frame);
void p_LIN_wait_us(uint32_t n);



//--------------------------------------------------------------
// init der LIN-Schnittstelle (im Slave-Mode)
//--------------------------------------------------------------
void UB_LIN_Slave_Init(void)
{
  // init aller Variabeln
  LIN_SLAVE.mode=WAIT_4_BREAK;
  LIN_SLAVE.data_ptr=0;
  LIN_SLAVE.crc=0;

  LIN_FRAME.frame_id=0;
  LIN_FRAME.data_len=0;
  LIN_FRAME.data[0]=0;

  // init der GPIOs
  p_LIN_initGPIO();
  // init vom UART
  p_LIN_initUART();
  // init vom NVIC
  p_LIN_initNVIC();

  // kurz warten bis Transceiver ready
  p_LIN_wait_us(LIN_POWERON_DELAY);

  // LIN Transceiver einschalten
  p_LIN_aktivateTransceiver();
}


//--------------------------------------------------------------
// prueft ob eine ID empfangen wurde
// (diese Funktion muss gepollt werden)
//
// falls "LIN_OK" muss die ID ausgewertet werden und
// sofort danach falls notwendig entweder
// "UB_LIN_SendData()" oder "UB_LIN_ReceiveData()"
// aufgerufen werden (sonst geht der Frame verloren)
//
// Return_wert :
//  LIN_OK       = ID wurde empfangen -> "frame.frame_id"
//  LIN_NO_ID    = es wurde keine ID empfangen
//--------------------------------------------------------------
LIN_ERR_t UB_LIN_checkID(LIN_FRAME_t *frame)
{
  LIN_ERR_t ret_wert=LIN_NO_ID;

  if(LIN_SLAVE.mode==ID_OK) {
    // ID-Field wurde empfangen
    LIN_SLAVE.mode=WAIT_4_CMD;
    frame->frame_id=LIN_FRAME.frame_id;
    ret_wert=LIN_OK;
  }
  else {
    // es wurde keine ID empfangen
    ret_wert=LIN_NO_ID;
  }

  return(ret_wert);
}


//--------------------------------------------------------------
// sendet Daten per LIN-Schnittstelle
// (nachdem eine ID empfangen wurde)
// frame : 
//   data_len = Anzahl der Daten die gesendet werden sollen
//   data[]   = Daten die gesendet werden sollen
//
// Return_wert :
//  LIN_OK         = Frame wurde gesendet
//  LIN_NO_ID      = es wurde keine ID empfangen
//  LIN_WRONG_LEN  = falsche Datenanzahl
//--------------------------------------------------------------
LIN_ERR_t UB_LIN_SendData(LIN_FRAME_t *frame)
{
  uint8_t checksum,n;

  // check vom mode
  if(LIN_SLAVE.mode!=WAIT_4_CMD) return(LIN_NO_ID);

  // check der laenge
  if((frame->data_len<1) || (frame->data_len>LIN_MAX_DATA)) { 
    LIN_SLAVE.mode=WAIT_4_BREAK;
    return(LIN_WRONG_LEN);
  }  
  
  // checksumme ausrechnen
  checksum=p_LIN_makeChecksum(frame);

  // warte bis letztes Byte gesendet wurde
  while (USART_GetFlagStatus(LIN_UART, USART_FLAG_TXE) == RESET);    

  // kleine Pause
  p_LIN_wait_us(LIN_FRAME_RESPONSE_DELAY); 

  // Daten senden
  LIN_SLAVE.mode=SEND_DATA;

  //------------------------
  // Data-Field [1...n]
  //------------------------
  for(n=0;n<frame->data_len;n++) {
    USART_SendData(LIN_UART, frame->data[n]);
    // warten bis DataField gesendet wurde
    while (USART_GetFlagStatus(LIN_UART, USART_FLAG_TC) == RESET);

    // kleine Pause
    p_LIN_wait_us(LIN_DATA_BYTE_DELAY);
  }  

  //------------------------
  // CRC-Field
  //------------------------
  USART_SendData(LIN_UART, checksum);
  // warten bis CRCField gesendet wurde
  while (USART_GetFlagStatus(LIN_UART, USART_FLAG_TC) == RESET); 

  // Ende vom Frame
  LIN_SLAVE.mode=WAIT_4_BREAK;

  return(LIN_OK);
}


//--------------------------------------------------------------
// empfaengt Daten per LIN-Schnittstelle
// (nachdem eine ID empfangen wurde)
// frame : 
//   data_len = Anzahl der Daten die empfangen werden sollen
// return :
//   data[]   = Daten die empfangen wurden (falls LIN_OK)
//
// Return_wert :
//  LIN_OK        = Frame wurde empfangen
//  LIN_NO_ID     = es wurde keine ID empfangen
//  LIN_WRONG_LEN = falsche Datenanzahl
//  LIN_RX_EMPTY  = kein Frame empfangen
//  LIN_WRONG_CRC = Checksumme falsch
//--------------------------------------------------------------
LIN_ERR_t UB_LIN_ReceiveData(LIN_FRAME_t *frame)
{
  uint32_t rx_timeout;
  uint8_t checksum,n;

  // check vom mode
  if(LIN_SLAVE.mode!=WAIT_4_CMD) return(LIN_NO_ID);

  // check der laenge 
  if((frame->data_len<1) || (frame->data_len>LIN_MAX_DATA)) { 
    LIN_SLAVE.mode=WAIT_4_BREAK;
    return(LIN_WRONG_LEN);
  }

  //-------------------------------
  // Slave vorbereiten
  //-------------------------------
  LIN_SLAVE.mode=RECEIVE_DATA;
  LIN_SLAVE.data_ptr=0;
  LIN_SLAVE.crc=0;

  LIN_FRAME.data_len=frame->data_len;
  LIN_FRAME.data[0]=0;

  //-------------------------------
  // warten bis Frame empfangen
  // oder Timeout
  //-------------------------------
  rx_timeout=0;
  n=0;
  do {   
    // timeout counter
    rx_timeout++;    
    if(rx_timeout>LIN_RX_TIMEOUT_CNT) {
      // schleife verlassen
      break;
    }
    // timeout zuruecksetzen, bei datenempfang
    if(LIN_SLAVE.data_ptr!=n) {
      n=LIN_SLAVE.data_ptr;
      rx_timeout=0;
    }
  }while(LIN_SLAVE.mode!=WAIT_4_BREAK);

  //-------------------------------
  // check ob frame empfangen wurden
  //-------------------------------
  if(LIN_SLAVE.mode!=WAIT_4_BREAK) {
    // kein Frame empfangen
    LIN_SLAVE.mode=WAIT_4_BREAK;
    return(LIN_RX_EMPTY);
  }   

  //-------------------------------
  // empfangene daten kopieren
  //-------------------------------
  for(n=0;n<frame->data_len;n++) {
    frame->data[n]=LIN_FRAME.data[n]; 
  }
  // checksumme ausrechnen
  checksum=p_LIN_makeChecksum(frame); 

  //-------------------------------
  // check ob crc ok
  //-------------------------------
  if(LIN_SLAVE.crc!=checksum) {
    // checksumme falsch
    return(LIN_WRONG_CRC);
  }

  //-------------------------------
  // daten sind ok
  //-------------------------------

  return(LIN_OK); 
}


//--------------------------------------------------------------
// interne Funktion
// init der GPIOs
//--------------------------------------------------------------
void p_LIN_initGPIO(void)
{
  GPIO_InitTypeDef GPIO_InitStructure;

  //-----------------------
  // NSLP vom Transceiver
  //-----------------------

  // Clock Enable
  RCC_AHB1PeriphClockCmd(LIN_NSLP_CLOCK, ENABLE);

  // Config als Digital-Ausgang
  GPIO_InitStructure.GPIO_Pin = LIN_NSLP_PIN;
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
  GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
  GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
  GPIO_Init(LIN_NSLP_PORT, &GPIO_InitStructure);

  // NSLP-Pin auf Lo-Pegel
  LIN_NSLP_PORT->BSRRH = LIN_NSLP_PIN;

  //-----------------------
  // RX + TX
  //-----------------------

  // Clock enable
  RCC_AHB1PeriphClockCmd(LIN_TX_CLOCK, ENABLE); // TX
  RCC_AHB1PeriphClockCmd(LIN_RX_CLOCK, ENABLE); // RX

  // GPIO-PINs als Alternative-Funktion mit PushPull
  GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF;
  GPIO_InitStructure.GPIO_Speed = GPIO_Speed_100MHz;
  GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
  GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
  // TX
  GPIO_InitStructure.GPIO_Pin = LIN_TX_PIN;
  GPIO_Init(LIN_TX_PORT, &GPIO_InitStructure);
  // RX
  GPIO_InitStructure.GPIO_Pin = LIN_RX_PIN;
  GPIO_Init(LIN_RX_PORT, &GPIO_InitStructure);

  // Alternative-Funktions mit den IO-Pins verbinden
  GPIO_PinAFConfig(LIN_TX_PORT,LIN_TX_SOURCE,LIN_UART_AF);
  GPIO_PinAFConfig(LIN_RX_PORT,LIN_RX_SOURCE,LIN_UART_AF); 
}


//--------------------------------------------------------------
// interne Funktion
// init vom UART
//--------------------------------------------------------------
void p_LIN_initUART(void)
{
  USART_InitTypeDef USART_InitStructure;

  // Clock enable
  if((LIN_UART==USART1) || (LIN_UART==USART6)) {
    RCC_APB2PeriphClockCmd(LIN_UART_CLOCK, ENABLE);
  }
  else {
    RCC_APB1PeriphClockCmd(LIN_UART_CLOCK, ENABLE);
  }

  // Oversampling bei LIN nicht erlaubt (OVER8=0) 

  // init mit Baudrate, 8Databits, 1Stopbit, keine Paritt, kein RTS+CTS
  USART_InitStructure.USART_BaudRate = LIN_UART_BAUD;
  USART_InitStructure.USART_WordLength = USART_WordLength_8b;
  USART_InitStructure.USART_StopBits = USART_StopBits_1;
  USART_InitStructure.USART_Parity = USART_Parity_No;
  USART_InitStructure.USART_HardwareFlowControl = USART_HardwareFlowControl_None;
  USART_InitStructure.USART_Mode = USART_Mode_Rx | USART_Mode_Tx;
  USART_Init(LIN_UART, &USART_InitStructure); 

  // UART enable
  USART_Cmd(LIN_UART, ENABLE); 

  // BreakDetection (10bit)
  USART_LINBreakDetectLengthConfig(LIN_UART, USART_LINBreakDetectLength_10b);

  // LIN enable
  USART_LINCmd(LIN_UART, ENABLE);
}


//--------------------------------------------------------------
// interne Funktion
// init vom NVIC
//--------------------------------------------------------------
void p_LIN_initNVIC(void)
{
  NVIC_InitTypeDef NVIC_InitStructure;

  // RX und BreakDetect-Interrupt enable
  USART_ITConfig(LIN_UART, USART_IT_RXNE, ENABLE);
  USART_ITConfig(LIN_UART, USART_IT_LBD, ENABLE);

  // enable Interrupt-Vector
  NVIC_InitStructure.NVIC_IRQChannel = LIN_UART_IRQ;
  NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 0;
  NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0;
  NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
  NVIC_Init(&NVIC_InitStructure);
}


//--------------------------------------------------------------
// interne Funktion
// LIN Transceiver (MAX13020)
// vom "Sleep-Mode" in "Normal-Slope-Mode" schalten
//--------------------------------------------------------------
void p_LIN_aktivateTransceiver(void)
{
  // NSLP-Pin auf Lo-Pegel
  LIN_NSLP_PORT->BSRRH = LIN_NSLP_PIN; // in "Sleep-Mode"
  // kurz warten (min. 10us)
  p_LIN_wait_us(LIN_AKTIV_DELAY);
  // NSLP-Pin auf Hi-Pegel  
  LIN_NSLP_PORT->BSRRL = LIN_NSLP_PIN; // in "Normal-Slope-Mode"
  // kurz warten (min. 10us)
  p_LIN_wait_us(LIN_AKTIV_DELAY);
}


//--------------------------------------------------------------
// interne Funktion
// checksumme ueber alle Daten berechnen
// (classic-Mode = invertierte Modulo256 Summe)
//
// ret_wert = checksumme
//--------------------------------------------------------------
uint8_t p_LIN_makeChecksum(LIN_FRAME_t *frame)
{
  uint8_t ret_wert=0,n;
  uint16_t dummy;

  // checksumme ausrechnen  
  dummy=0;
  for(n=0;n<frame->data_len;n++) {
    dummy+=frame->data[n];
    if(dummy>0xFF) {
      dummy-=0xFF;
    } 
  }
  ret_wert=(uint8_t)(dummy);
  ret_wert^=0xFF;
  
  return(ret_wert);
}


//--------------------------------------------------------------
// interne Funktion
// kleine Pause (ohne Timer)
// 1   = ca. 1us
// 10  = ca. 10us
// 100 = ca. 100us
//--------------------------------------------------------------
void p_LIN_wait_us(uint32_t n)
{
  volatile uint32_t p,t;

  // kleine Pause
  for(p=0;p<n;p++) {
    for(t=0;t<15;t++); // ca 1us
  }
}


//--------------------------------------------------------------
// interne Funktion
// UART ISR
// wird aufgerufen wenn ein LIN-Field oder ein Break-Field
// empfangen wurde
// (auch alle Slave TX-Fields werden als RX wieder empfangen)
//--------------------------------------------------------------
void LIN_UART_ISR(void) {
  uint16_t wert;

  if (USART_GetITStatus(LIN_UART, USART_IT_LBD) == SET) {
    //---------------------------
    // BreakField
    //---------------------------
    // flag loeschen
    USART_ClearITPendingBit(LIN_UART, USART_IT_LBD);

    LIN_SLAVE.mode=RECEIVE_SYNC;
  }


  if (USART_GetITStatus(LIN_UART, USART_IT_RXNE) == SET) {
    // Daten auslesen
    wert=USART_ReceiveData(LIN_UART);

    // check welcher Mode gerade aktiv ist
    if(LIN_SLAVE.mode==RECEIVE_SYNC) {
      //---------------------------
      // SyncField
      //---------------------------
      if(wert==LIN_SYNC_DATA) {
        LIN_SLAVE.mode=RECEIVE_ID;
      }
      else {
        LIN_SLAVE.mode=WAIT_4_BREAK;              
      }
    }
    else if(LIN_SLAVE.mode==RECEIVE_ID) {
      //---------------------------
      // IDField
      //---------------------------
      LIN_SLAVE.mode=ID_OK;
      LIN_FRAME.frame_id=(uint8_t)(wert);      
    }
    else if(LIN_SLAVE.mode==RECEIVE_DATA) {
      //---------------------------
      // DataField
      //---------------------------
      // Daten speichern
      LIN_FRAME.data[LIN_SLAVE.data_ptr]=(uint8_t)(wert);
      // Pointer weiterschalten
      LIN_SLAVE.data_ptr++;
      // check ob alle Daten empfangen sind
      if(LIN_SLAVE.data_ptr>=LIN_FRAME.data_len) {          
        LIN_SLAVE.mode=RECEIVE_CRC;
      }
    }
    else if(LIN_SLAVE.mode==RECEIVE_CRC) { 
      //---------------------------
      // CRCField
      //---------------------------     
      LIN_SLAVE.crc=(uint8_t)(wert);
      LIN_SLAVE.mode=WAIT_4_BREAK; 
    }
  }
}
